"""Predicate."""

from typing import Any, Callable

from hamcrest import assert_that
from hamcrest.core.matcher import Matcher

from .status import Status
from .verification import Verification


class MatcherPredicate:
    """Predicate of a Hamcrest matcher."""

    def __init__(self, matcher: Matcher):
        self._matcher = matcher

    def __call__(self, actual: Any, *args: Any, **kwargs: Any) -> Verification:
        try:
            assert_that(actual, self._matcher)
        except AssertionError as error:
            message = str(error).strip()
            return Verification(status=Status.UNSTABLE, message=message)
        except Exception as error:
            return Verification.of_error(error)

        return Verification.succeed()


class DynamicMatcherPredicate:
    """Predicate of a dynamic Hamcrest matcher and conversion."""

    def __init__(
        self,
        matcher_factory: Callable,
        converter: Callable[[Any], Any] = lambda x: x,
    ):
        self._matcher_factory = matcher_factory
        self._converter = converter

    def __call__(self, actual: Any, *args: Any, **kwargs: Any) -> Verification:
        try:
            matcher = self._matcher_factory(*args, **kwargs)
            predicated_value = self._converter(actual)
        except Exception as error:
            return Verification.of_error(error)

        predicate = MatcherPredicate(matcher)
        return predicate(predicated_value, *args, **kwargs)



"""Datetime compilations."""

import re
from datetime import timedelta

from .error import CompilationError


RELATIVE_DATETIME_PATTERN = re.compile(
    r'([+\-]?\d+)\s*(day|hour|minute|second)s?'
)


def compile_timedelta(value: str) -> timedelta:
    match = RELATIVE_DATETIME_PATTERN.search(value.lower())
    if not match:
        raise CompilationError(f'Invalid datetime format: {value}')
    offset = int(match.group(1))
    unit = match.group(2) + 's'
    return timedelta(**{unit: offset})



"""Scenario"""

from typing import Iterator, List, Optional

from .case import Case


class Scenario:
    """
    When given no cases, then skips.
    >>> scenario = Scenario()
    >>> scenario.label
    >>> list(scenario.cases())
    []

    When given a label, then returns it.
    >>> scenario = Scenario(label='label')
    >>> scenario.label
    'label'

    When given cases, then iterates them.
    >>> from unittest.mock import sentinel
    >>> scenario = Scenario(cases=[sentinel.case1, sentinel.case2])
    >>> cases = scenario.cases()
    >>> next(cases)
    sentinel.case1
    >>> next(cases)
    sentinel.case2
    """
    def __init__(self, label: Optional[str] = None, cases: List[Case] = []):
        self._label = label
        self._cases = cases

    @property
    def label(self) -> Optional[str]:
        return self._label

    def cases(self) -> Iterator[Case]:
        return iter(self._cases)



"""Response descriptions."""

import json
from dataclasses import dataclass
from typing import Any, List

from .description import Description, Predicate
from .status import Status, merge_statuses
from .verification import Verification


@dataclass(frozen=True)
class ResponseVerification:
    status: Status
    status_code: Verification
    body: Verification


class ResponseDescription:

    def __init__(
        self,
        status_code_predicates: List[Predicate],
        body_descriptions: List[Description],
    ):
        self._status_code_predicates = status_code_predicates
        self._body_descriptions = body_descriptions

    def __call__(
        self,
        status_code: int,
        body: str,
        *args: Any,
        **kwargs: Any,
    ) -> ResponseVerification:
        """`*args` and `**kwargs` will be delegated to descriptions."""

        status_code_verification = self._verify_status_code(
            status_code,
            *args,
            **kwargs,
        )
        try:
            body_verification = self._verify_body(body, *args, **kwargs)
        except Exception as error:
            body_verification = Verification.of_error(error)

        status = merge_statuses(
            status_code_verification.status,
            body_verification.status,
        )
        return ResponseVerification(
            status=status,
            status_code=status_code_verification,
            body=body_verification,
        )

    @property
    def status_code_predicates(self) -> List[Predicate]:
        return self._status_code_predicates

    @property
    def body_descriptions(self) -> List[Description]:
        return self._body_descriptions

    def _verify_status_code(
        self,
        code: int,
        *args: Any,
        **kwargs: Any,
    ) -> Verification:
        children = [
            predicate(code, *args, **kwargs)
            for predicate in self._status_code_predicates
        ]
        status = merge_statuses(v.status for v in children)
        return Verification(status=status, children=children)

    def _verify_body(
        self,
        body: str,
        *args: Any,
        **kwargs: Any,
    ) -> Verification:
        if not self._body_descriptions:
            return Verification.skipped()

        data = json.loads(body)
        verifications = [
            describe(data, *args, **kwargs)
            for describe in self._body_descriptions
        ]
        status = merge_statuses(v.status for v in verifications)
        return Verification(status=status, children=verifications)



"""Description compilation."""

from collections.abc import Mapping
from typing import Optional

from preacher.core.description import Description
from .error import CompilationError
from .predicate import PredicateCompiler
from .extraction import ExtractionCompiler
from .util import run_on_key, map_on_key


_KEY_DESCRIBE = 'describe'
_KEY_SHOULD = 'should'


class DescriptionCompiler:

    def __init__(
        self,
        extraction_compiler: Optional[ExtractionCompiler] = None,
        predicate_compiler: Optional[PredicateCompiler] = None,
    ):
        self._extraction_compiler = extraction_compiler or ExtractionCompiler()
        self._predicate_compiler = predicate_compiler or PredicateCompiler()

    def compile(self, obj: Mapping):
        extraction_obj = obj.get(_KEY_DESCRIBE)
        if (
            not isinstance(extraction_obj, Mapping)
            and not isinstance(extraction_obj, str)
        ):
            raise CompilationError(
                message='Description.describe must be a mapping or a string',
                path=[_KEY_DESCRIBE],
            )
        extraction = run_on_key(
            _KEY_DESCRIBE,
            self._extraction_compiler.compile,
            extraction_obj
        )

        predicate_objs = obj.get(_KEY_SHOULD, [])
        if not isinstance(predicate_objs, list):
            predicate_objs = [predicate_objs]
        predicates = list(map_on_key(
            _KEY_SHOULD,
            self._predicate_compiler.compile,
            predicate_objs,
        ))

        return Description(extraction=extraction, predicates=predicates)



"""Request compilation."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Optional, Union

from preacher.core.request import Request
from .error import CompilationError
from .util import or_default


_KEY_PATH = 'path'
_KEY_PARAMS = 'params'


@dataclass(frozen=True)
class _Compiled:
    path: Optional[str] = None
    params: Optional[Mapping] = None

    def to_request(
        self: _Compiled,
        default_path: str,
        default_params: Mapping,
    ) -> Request:
        return Request(
            path=or_default(self.path, default_path),
            params=or_default(self.params, default_params),
        )


def _compile(obj: Union[Mapping, str]) -> _Compiled:
    if isinstance(obj, str):
        return _compile({_KEY_PATH: obj})

    path = obj.get(_KEY_PATH)
    if path is not None and not isinstance(path, str):
        raise CompilationError('Must be a string', path=[_KEY_PATH])

    params = obj.get(_KEY_PARAMS)
    if params is not None and not isinstance(params, Mapping):
        raise CompilationError('Must be a mapping', path=[_KEY_PARAMS])

    return _Compiled(path=path, params=params)


class RequestCompiler:

    def __init__(self, path: str = '', params: Mapping = {}):
        self._path = path
        self._params = params

    def compile(self, obj: Union[Mapping, str]) -> Request:
        compiled = _compile(obj)
        return compiled.to_request(
            default_path=self._path,
            default_params=self._params,
        )

    def of_default(self, obj: Union[Mapping, str]) -> RequestCompiler:
        compiled = _compile(obj)
        return RequestCompiler(
            path=or_default(compiled.path, self._path),
            params=or_default(compiled.params, self._params),
        )



"""Predicate compilation."""

from collections.abc import Mapping
from datetime import datetime
from typing import Any, Callable

from hamcrest import greater_than, less_than
from hamcrest.core.matcher import Matcher

from preacher.core.description import Predicate
from preacher.core.datetime import now, parse_datetime
from preacher.core.predicate import MatcherPredicate, DynamicMatcherPredicate
from .datetime import compile_timedelta
from .error import CompilationError
from .matcher import compile as compile_matcher
from .util import run_on_key


PREDICATE_MAP = {
    'be_before': lambda value:
        _compile_datetime_predicate('be_before', value, less_than),
    'be_after': lambda value:
        _compile_datetime_predicate('be_after', value, greater_than),
}


class PredicateCompiler:

    def compile(self, obj: Any) -> Predicate:
        if isinstance(obj, Mapping):
            if len(obj) != 1:
                raise CompilationError(
                    f'Must have only 1 element, but has {len(obj)}'
                )

            key, value = next(iter(obj.items()))
            if key in PREDICATE_MAP:
                return PREDICATE_MAP[key](value)

        matcher = compile_matcher(obj)
        return MatcherPredicate(matcher)


def _compile_datetime_predicate(
    key: str,
    obj: Any,
    matcher_func: Callable[[datetime], Matcher],
) -> DynamicMatcherPredicate:
    if not isinstance(obj, str):
        raise CompilationError(message=f'Must be a string', path=[key])

    delta = run_on_key(key, compile_timedelta, obj)

    def _matcher_factory(*args: Any, **kwargs: Any) -> Matcher:
        origin = kwargs.get('request_datetime') or now()
        return matcher_func(origin + delta)

    return DynamicMatcherPredicate(
        matcher_factory=_matcher_factory,
        converter=parse_datetime,
    )



"""Scenario running helpers."""

from dataclasses import dataclass
from typing import List, Optional

from .case import CaseResult
from .scenario import Scenario
from .status import Status, merge_statuses


@dataclass(frozen=True)
class ScenarioResult:
    label: Optional[str]
    status: Status
    case_results: List[CaseResult]


def run_scenario(
    scenario: Scenario,
    base_url: str,
    retry: int = 0,
) -> ScenarioResult:
    case_results = [
        case(base_url=base_url, retry=retry)
        for case in scenario.cases()
    ]
    status = merge_statuses(result.status for result in case_results)
    return ScenarioResult(
        label=scenario.label,
        status=status,
        case_results=case_results,
    )



"""Matcher compilation."""

from collections.abc import Mapping
from typing import Any

import hamcrest
from hamcrest.core.matcher import Matcher

from .error import CompilationError
from .util import run_on_key


_STATIC_MATCHER_MAP = {
    # For objects.
    'be_null': hamcrest.is_(hamcrest.none()),
    'not_be_null': hamcrest.is_(hamcrest.not_none()),

    # For collections.
    'be_empty': hamcrest.is_(hamcrest.empty()),
}

_MATCHER_FUNCTION_MAP_TAKING_SINGLE_VALUE = {
    # For objects.
    'equal': lambda expected: hamcrest.is_(hamcrest.equal_to(expected)),
    'have_length': hamcrest.has_length,

    # For numbers.
    'be_greater_than': (
        lambda value: hamcrest.is_(hamcrest.greater_than(value))
    ),
    'be_greater_than_or_equal_to': (
        lambda value: hamcrest.is_(hamcrest.greater_than_or_equal_to(value))
    ),
    'be_less_than': (
        lambda value: hamcrest.is_(hamcrest.less_than(value))
    ),
    'be_less_than_or_equal_to': (
        lambda value: hamcrest.is_(hamcrest.less_than_or_equal_to(value))
    ),

    # For strings.
    'contain_string': hamcrest.contains_string,
    'start_with': hamcrest.starts_with,
    'end_with': hamcrest.ends_with,
    'match_regexp': hamcrest.matches_regexp,
}

_MATCHER_FUNCTION_MAP_TAKING_SINGLE_MATCHER = {
    'be': hamcrest.is_,
    'not': hamcrest.not_,
    'have_item': hamcrest.has_item,
}


def _compile_taking_single_matcher(key: str, value: Any):
    func = _MATCHER_FUNCTION_MAP_TAKING_SINGLE_MATCHER[key]

    if isinstance(value, str) or isinstance(value, Mapping):
        inner = run_on_key(key, compile, value)
    else:
        inner = hamcrest.equal_to(value)

    return func(inner)


def compile(obj: Any) -> Matcher:
    if isinstance(obj, str):
        if obj in _STATIC_MATCHER_MAP:
            return _STATIC_MATCHER_MAP[obj]

    if isinstance(obj, Mapping):
        if len(obj) != 1:
            raise CompilationError(
                f'Must have only 1 element, but has {len(obj)}'
            )

        key, value = next(iter(obj.items()))

        if key in _MATCHER_FUNCTION_MAP_TAKING_SINGLE_VALUE:
            return _MATCHER_FUNCTION_MAP_TAKING_SINGLE_VALUE[key](value)

        if key in _MATCHER_FUNCTION_MAP_TAKING_SINGLE_MATCHER:
            return _compile_taking_single_matcher(key, value)

    return hamcrest.equal_to(obj)



"""Test case."""

from dataclasses import dataclass
from typing import Optional

from .request import Request
from .response_description import (
    ResponseDescription,
    ResponseVerification,
)
from .status import Status, merge_statuses
from .verification import Verification


@dataclass(frozen=True)
class CaseResult:
    status: Status
    request: Verification
    response: Optional[ResponseVerification] = None
    label: Optional[str] = None


class Case:

    def __init__(
        self,
        request: Request,
        response_description: ResponseDescription,
        label: Optional[str] = None,
    ):
        self._label = label
        self._request = request
        self._response_description = response_description

    def __call__(self, base_url: str, retry: int = 0) -> CaseResult:
        if retry < 0:
            raise ValueError(
                f'Retry count must be positive or 0, given {retry}'
            )
        for _ in range(1 + retry):
            result = self._run(base_url)
            if result.status.is_succeeded:
                break

        return result

    def _run(self, base_url: str) -> CaseResult:
        try:
            response = self._request(base_url)
        except Exception as error:
            return CaseResult(
                status=Status.FAILURE,
                request=Verification.of_error(error),
            )
        request_verification = Verification.succeed()

        response_verification = self._response_description(
            status_code=response.status_code,
            body=response.body,
            request_datetime=response.request_datetime,
        )

        status = merge_statuses(
            request_verification.status,
            response_verification.status,
        )
        return CaseResult(
            status=status,
            request=request_verification,
            response=response_verification,
            label=self._label,
        )

    @property
    def label(self) -> Optional[str]:
        return self._label

    @property
    def request(self) -> Request:
        return self._request

    @property
    def response_description(self) -> ResponseDescription:
        return self._response_description



"""Response description compilations."""

from collections.abc import Mapping
from typing import Any, Optional

from preacher.core.description import Description
from preacher.core.response_description import ResponseDescription
from .error import CompilationError
from .description import DescriptionCompiler
from .predicate import PredicateCompiler
from .util import map_on_key


_KEY_STATUS_CODE = 'status_code'
_KEY_BODY = 'body'


class ResponseDescriptionCompiler:

    def __init__(
        self,
        predicate_compiler: Optional[PredicateCompiler] = None,
        description_compiler: Optional[DescriptionCompiler] = None,
    ):
        self._predicate_compiler = predicate_compiler or PredicateCompiler()
        self._description_compiler = (
            description_compiler
            or DescriptionCompiler(
                predicate_compiler=self._predicate_compiler
            )
        )

    def compile(self, obj: Mapping) -> ResponseDescription:
        status_code_predicate_objs = obj.get(_KEY_STATUS_CODE, [])
        if not isinstance(status_code_predicate_objs, list):
            status_code_predicate_objs = [status_code_predicate_objs]
        status_code_predicates = list(map_on_key(
            key=_KEY_STATUS_CODE,
            func=self._predicate_compiler.compile,
            items=status_code_predicate_objs,
        ))

        body_description_objs = obj.get(_KEY_BODY, [])
        if isinstance(body_description_objs, Mapping):
            body_description_objs = [body_description_objs]
        if not isinstance(body_description_objs, list):
            raise CompilationError(
                message='ResponseDescription.body must be a list or a mapping',
                path=[_KEY_BODY],
            )
        body_descriptions = list(map_on_key(
            key=_KEY_BODY,
            func=self._compile_description,
            items=body_description_objs,
        ))

        return ResponseDescription(
            status_code_predicates=status_code_predicates,
            body_descriptions=body_descriptions,
        )

    def _compile_description(self, obj: Any) -> Description:
        if not isinstance(obj, Mapping):
            raise CompilationError('Description must be a mapping')
        return self._description_compiler.compile(obj)



"""Description."""

from typing import Any, Callable, List

from .status import merge_statuses
from .verification import Verification

Extraction = Callable[[Any], Any]
Predicate = Callable


class Description:

    def __init__(self, extraction: Extraction, predicates: List[Predicate]):
        self._extraction = extraction
        self._predicates = predicates

    def __call__(self, value: Any, *args: Any, **kwargs: Any) -> Verification:
        """`*args` and `**kwargs` will be delegated to predicates."""

        try:
            verified_value = self._extraction(value)
        except Exception as error:
            return Verification.of_error(error)

        verifications = [
            predicate(verified_value, *args, **kwargs)
            for predicate in self._predicates
        ]
        status = merge_statuses(v.status for v in verifications)
        return Verification(status, children=verifications)

    @property
    def extraction(self) -> Extraction:
        return self._extraction

    @property
    def predicates(self) -> List[Predicate]:
        return self._predicates



"""Compilation errors."""

from __future__ import annotations

from typing import List, Optional


class CompilationError(Exception):
    """Compilation errors."""

    def __init__(
        self,
        message: str,
        path: List[str] = [],
        cause: Optional[Exception] = None,
    ):
        super().__init__(message)
        self._message = message
        self._path = path
        self._cause = cause

    @property
    def path(self) -> List[str]:
        return self._path

    def of_parent(self, parent_path: List[str]) -> CompilationError:
        return CompilationError(
            message=self._message,
            path=parent_path + self._path,
            cause=self._cause,
        )

    def __str__(self):
        message = super().__str__()
        if not self._path:
            return message

        path = '.'.join(self._path)
        return f'{message}: {path}'



"""Request."""

from collections.abc import Mapping
from dataclasses import dataclass
from datetime import datetime

import requests

from .datetime import now

from preacher import __version__ as _VERSION


_DEFAULT_HEADERS = {
    'User-Agent': f'Preacher {_VERSION}'
}


@dataclass(frozen=True)
class Response:
    status_code: int
    headers: Mapping
    body: str
    request_datetime: datetime


class Request:
    def __init__(self, path: str, params: Mapping):
        self._path = path
        self._params = params

    def __call__(self, base_url: str) -> Response:
        request_datetime = now()

        res = requests.get(
            base_url + self._path,
            headers=_DEFAULT_HEADERS,
            params=self._params,
        )

        return Response(
            status_code=res.status_code,
            headers=res.headers,
            body=res.text,
            request_datetime=request_datetime
        )

    @property
    def path(self) -> str:
        return self._path

    @property
    def params(self) -> Mapping:
        return self._params



"""Preacher: A Web API Verification Tool."""

__version__ = '0.9.7'
__author__ = 'Yu Mochizuki'
__author_email__ = 'ymoch.dev@gmail.com'

NAME = 'preacher'
DESCRIPTION = 'A Web API verification tool.'
URL = 'https://github.com/ymoch/preacher'
LICENSE = 'MIT'


def __dummy__() -> None:
    """
    Dummy function so that tests cover this module.
    >>> __dummy__()
    """
    pass



"""Datetime utilities for Preacher core."""

import datetime
import time


def now() -> datetime.datetime:
    return datetime.datetime.now(_system_timezone())


def parse_datetime(value: str) -> datetime.datetime:
    """
    Only takes ISO 6801 expanded format.
    """
    if value.endswith('Z'):
        value = value[:-1] + '+00:00'
    return datetime.datetime.fromisoformat(value)


def _system_timezone() -> datetime.timezone:
    localtime = time.localtime()
    return datetime.timezone(
        offset=datetime.timedelta(seconds=localtime.tm_gmtoff),
        name=localtime.tm_zone,
    )



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['preacher',
 'preacher.app',
 'preacher.app.cli',
 'preacher.compilation',
 'preacher.core',
 'preacher.presentation']

package_data = \
{'': ['*']}

install_requires = \
['pyhamcrest>=1.9,<2.0',
 'pyjq>=2.3,<3.0',
 'requests>=2.21,<3.0',
 'ruamel.yaml>=0.15.89,<0.16.0']

entry_points = \
{'console_scripts': ['preacher-cli = preacher.app.cli.main:main']}

setup_kwargs = {
    'name': 'preacher',
    'version': '0.9.7',
    'description': 'A web API verification tool.',
    'long_description': '========\nPreacher\n========\n\n.. image:: https://img.shields.io/badge/python-3.7+-blue.svg\n    :target: https://www.python.org/\n.. image:: https://badge.fury.io/py/preacher.svg\n    :target: https://badge.fury.io/py/preacher\n.. image:: https://travis-ci.org/ymoch/preacher.svg?branch=master\n    :target: https://travis-ci.org/ymoch/preacher\n.. image:: https://codecov.io/gh/ymoch/preacher/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/ymoch/preacher\n.. image:: https://img.shields.io/lgtm/grade/python/g/ymoch/preacher.svg\n    :target: https://lgtm.com/projects/g/ymoch/preacher/context:python\n\nPreacher verifies API servers,\nwhich requests to API servers and verify the response along to given scenarios.\n\nScenarios are written in `YAML`_ and `jq`_ queries\nso that any developers can write without learning toughly.\n\n\nUsage\n=====\n\nRequirements\n------------\nSupports only Python 3.7+.\n\n.. code-block:: sh\n\n    pip install preacher\n    preacher-cli scenario.yml\n\n    # Help command is available.\n    preacher-cli --help\n\n\nWriting Your Own Scenarios\n==========================\n\nExample\n-------\nHere is a simple configuration example.\n\n.. code-block:: yaml\n\n    label: Scenario example\n    default:\n      request:\n        path: /path\n    cases:\n      - label: Simple\n        request: /path/to/foo\n        response:\n          status_code: 200\n          body:\n            - describe: .foo\n              should:\n                equal: bar\n      - label: A Little Complecated\n        request:\n          path: /path/to/foo\n          params:\n            key1: value\n            key2:\n              - value1\n              - value2\n        response:\n          status_code:\n            - be_greater_than_or_equal_to: 200\n            - be_less_than: 400\n          body:\n            - describe:\n                jq: .foo\n              should:\n                - start_with: x\n                - end_with: y\n\nGrammer\n-------\n\nScenario\n********\nA ``Scenario`` is written in `YAML`_.\nA ``Scenario`` is a mapping that consists of below:\n\n- label: ``String`` (Recommended)\n    - A label of this scenario.\n    - This field is actually optional but recommended to tell this scenario from another.\n- default: ``Default`` (Optional)\n    - Default of this scenario.\n- cases: ``List<Case>``\n    - Test cases.\n\nCase\n****\nA ``Case`` is a mapping that consists of below:\n\n- label: ``String`` (Recommended)\n    - A label of this case.\n    - This field is actually optional but recommended to tell this case from another.\n- request: ``Request`` (Optional)\n    - A request.\n- response: ``ResponseDescription`` (Optional)\n    - A response description.\n\nRequest\n*******\nA ``Request`` is a mapping or a string.\n\nA mapping for ``Request`` has items below:\n\n- path: ``String`` (Optional)\n    - A request path. The default value is ``\'\'``.\n- params: ``Mapping<String, String>`` (Optional)\n    - Query parameters as a mapping of keys to values.\n\nWhen given a string, that is equivalent to ``{"path": it}``.\n\nResponse Decription\n*******************\nA ``ResponseDescription`` is a mapping that consists of below:\n\n- status_code: ``Integer``, ``Predicate`` or ``List<Predicate>`` (Optional)\n    - Predicates that match a status code as an integer value.\n    - When given a number, that is equivalent to ``{"equal": it}``.\n- body: ``Description`` or ``List<Description>`` (Optional)\n    - Descriptions that descript the response body.\n\nDescription\n***********\nA ``Description`` is a mapping that consists of below:\n\n- describe: ``String`` or ``Extraction``\n    - An extraction process.\n    - When given a string, that is equivalent to ``{"jq": it}``.\n- should: ``Predicate``, or ``List<Predicate>>`` (Optional)\n    - Predicates that match the extracted value.\n\nExtraction\n**********\nAn ``Extraction`` is a mapping or a string.\n\nA mapping for ``Extraction`` has one of below:\n\n- jq: ``String``\n    - A `jq`_ query.\n\nWhen fiven a string, that is equivalent to ``{"jq": it}``.\n\nPredicate\n*********\nA ``Predicate`` is a ``Matcher`` (can be extended in the future).\n\nMatcher\n*******\nA ``Matcher`` is a string or a mapping.\n\nAllowed strings are:\n\n- be_null\n- not_be_null\n- be_empty\n\nA mapping for ``Matcher`` has an item. Allowed items are:\n\n- be: ``Value`` or ``Matcher``\n    - Matches when it matches the given value or the given matcher.\n    - When given ``Value``, that is equivalent to ``{"equal": it}``.\n- not: ``Value`` or ``Matcher``\n    - Matches when it doesn\'t match the given value or the given matcher.\n    - When given ``Value``, that is equivalent to ``{"not": {"equal": it}}``\n- equal: ``Value``\n    - Matches when it equals the given value.\n- have_length: ``Integer``\n    - Matches when it has a length and its length is equal to the given value.\n- be_greater_than: ``Comparable``\n    - Matches when it is greater than the given value (it > argument).\n- be_greater_than_or_equal_to: ``Comparable``\n    - Matches when it is greater than or equal to the given value (it >= argument).\n- be_less_than: ``Comparable``\n    - Matches when it is less than the given value (it < argument).\n- be_less_than_or_equal_to: ``Comparable``\n    - Matches when it is less than or equal to the given value (it < argument).\n- contain_string: ``String``\n    - Matches when it is an string and contains the given value.\n- start_with: ``String``\n    - Matches when it is an string and starts with the given value.\n- end_with: ``String``\n    - Matches when it is an string and ends with the given value.\n- match_regexp: ``String``\n    - Matches when it is an string and matches the given regular expression.\n- have_item: ``Value`` or ``Matcher``\n    - Matches when it is a collection and has the given item.\n    - When given ``Value``, that is equivalent to ``{"equal": it}``.\n- be_before:\n    - Matches when it is a datetime and before the given datetime.\n    - Predicated values must be in extended ISO 8601 format\n      like ``2019-01-23T12:34:56Z``.\n    - When given an offset, then compares to the datetime when the request starts.\n        - Days, hours, minutes and seconds offsets are available.\n        - When given a positive offset like ``1 day`` or ``+2 hours``,\n          then compares to the future datetime.\n        - When given a negative offset like ``-1 minute`` or ``-2 seconds``,\n          then compares to the past datetime.\n- be_after:\n    - Matches when it is a datetime and after the given datetime.\n    - Usage is the same as ``be_before``.\n\nDefault\n*******\nA ``Default`` is a mapping that consists of below:\n\n- request: ``Request`` (Optional)\n    - A request to override the default request values.\n\n\n.. _YAML: https://yaml.org/\n.. _jq: https://stedolan.github.io/jq/\n.. _pipenv: https://pipenv.readthedocs.io/\n',
    'author': 'Yu Mochizuki',
    'author_email': 'ymoch.dev@gmail.com',
    'url': 'https://github.com/ymoch/preacher',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Preacher CLI."""

from __future__ import annotations

import argparse
import logging
import sys

from preacher import __version__ as VERSION
from preacher.presentation.logging import LoggingPresentation
from .application import Application


HANDLER = logging.StreamHandler()
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(HANDLER)


LOGGING_LEVEL_MAP = {
    'skipped': logging.DEBUG,
    'success': logging.INFO,
    'unstable': logging.WARN,
    'failure': logging.ERROR,
}


def zero_or_positive_int(value: str) -> int:
    int_value = int(value)
    if int_value < 0:
        raise argparse.ArgumentTypeError(
            f"must be positive or 0, given {int_value}"
        )
    return int_value


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'conf',
        nargs='+',
        help='config file paths'
    )
    parser.add_argument(
        '-v', '--version',
        action='version',
        version=VERSION,
    )
    parser.add_argument(
        '-u', '--url',
        metavar='url',
        help='specify the base URL',
        default='http://localhost:5000',
    )
    parser.add_argument(
        '-l', '--level',
        choices=LOGGING_LEVEL_MAP.keys(),
        help='show only above or equal to this level',
        default='success',
    )
    parser.add_argument(
        '-r', '--retry',
        type=zero_or_positive_int,
        help='max retry count',
        default=0,
    )
    parser.add_argument(
        '-c', '--scenario-concurrency',
        type=int,
        help='concurrency for scenarios',
        default=1,
    )

    return parser.parse_args()


def main() -> None:
    """Main."""
    args = parse_args()

    level = LOGGING_LEVEL_MAP[args.level]
    HANDLER.setLevel(level)
    LOGGER.setLevel(level)

    view = LoggingPresentation(LOGGER)
    base_url = args.url
    retry = args.retry
    app = Application(view=view, base_url=base_url, retry=retry)

    config_paths = args.conf
    scenario_concurrency = args.scenario_concurrency
    app.run_concurrently(config_paths, concurrency=scenario_concurrency)

    if not app.is_succeeded:
        sys.exit(1)



from multiprocessing import Pool
from typing import Callable, Iterable, Iterator

import ruamel.yaml as yaml

from preacher.core.scenario_running import ScenarioResult, run_scenario
from preacher.compilation.scenario import ScenarioCompiler
from preacher.presentation.logging import LoggingPresentation


MapFunction = Callable[
    [
        Callable[[str], ScenarioResult],
        Iterable[str]
    ],
    Iterator[ScenarioResult]
]


class Application:
    def __init__(
        self,
        view: LoggingPresentation,
        base_url: str,
        retry: int = 0,
    ):
        self._view = view
        self._base_url = base_url
        self._retry = retry

        self._scenario_compiler = ScenarioCompiler()
        self._is_succeeded = True

    @property
    def is_succeeded(self) -> bool:
        return self._is_succeeded

    def run(
        self,
        config_paths: Iterable[str],
        map_func: MapFunction = map,
    ) -> None:
        results = map_func(self._run_each, config_paths)
        for result in results:
            self._is_succeeded &= result.status.is_succeeded
            self._view.show_scenario_result(result)

    def run_concurrently(
        self,
        config_paths: Iterable[str],
        concurrency: int,
    ) -> None:
        with Pool(concurrency) as pool:
            self.run(config_paths, map_func=pool.imap)

    def _run_each(self, config_path: str) -> ScenarioResult:
        with open(config_path) as config_file:
            config = yaml.safe_load(config_file)
        scenario = self._scenario_compiler.compile(config)
        return run_scenario(
            scenario,
            base_url=self._base_url,
            retry=self._retry,
        )
